
// JavaScript classes for 2D geometry

// These next lines are for use with http://jshint.com/
// jshint esversion: 6
// jshint bitwise: false
// jshint laxbreak: true

"use strict";

// ============================================================
// Vec2 objects are for storing 2D points and 2D vectors.
// To create a new instance, use the new keyword:
//    let vector_a = new Vec2();
//    let vector_b = new Vec2(x,y);
// These objects are mutable. After an instance has been created,
// its coordinates can be changed by writing directly to data members:
//    vector_a.x = 10;
//    vector_a.y = -2;
// ============================================================
class Vec2 {
	constructor(x=0,y=0) {
		this.x = x;
		this.y = y;
	}

	// Performs a deep copy.
	copy(other) {
		this.x = other.x; this.y = other.y;
	}

	// Returns the negation of the vector.
	negate() { return new Vec2(-this.x,-this.y); }

	// Returns the Euclidean length (also called magnitude or L2-norm) of the vector.
	norm() { return Math.sqrt( this.x*this.x + this.y*this.y ); }

	// Returns the squared length.
	// This is useful when the caller needs to compare
	// the length of a vector to a pre-defined threshold,
	// or compare the lengths of two vectors:
	// in such cases, comparing the squared length is sufficient,
	// and saves a square root operation.
	normSquared() { return this.x*this.x + this.y*this.y; }

	// Returns a normalized vector of unit length.
	normalize() {
		let n = this.norm();
		if ( n > 0 ) {
			let k = 1.0/n;
			return new Vec2( k*this.x, k*this.y );
		}
		return new Vec2();
	}

	// A static method that returns the sum of two vectors.
	static sum( v1, v2 ) {
		return new Vec2( v1.x+v2.x, v1.y+v2.y );
	}

	// A static method that returns the difference of two vectors.
	static diff( v1, v2 ) {
		return new Vec2( v1.x-v2.x, v1.y-v2.y );
	}
	// A static method that returns the product of a vector with a scalar.
	static mult( v, k ) {
		return new Vec2( k*v.x, k*v.y );
	}
	// A static method that returns the dot product of two vectors
	static dot( v1, v2 ) {
		return v1.x*v2.x + v1.y*v2.y;
	}
	// A static method that returns the centroid of two vectors
	static average( v1, v2 ) {
		return new Vec2( (v1.x+v2.x)*0.5, (v1.y+v2.y)*0.5 );
	}

	static distance( v1, v2 ) {
		let dx = v1.x - v2.x;
		let dy = v1.y - v2.y;
		return Math.sqrt( dx*dx + dy*dy );
	}
	static distanceSquared( v1, v2 ) {
		let dx = v1.x - v2.x;
		let dy = v1.y - v2.y;
		return dx*dx + dy*dy;
	}

	// A static method that computes the centroid of an array of vectors
	static centroidOfPoints( points ) {
		let x = 0;
		let y = 0;
		for ( let i = 0; i < points.length; ++i ) {
			x += points[i].x;
			y += points[i].y;
		}
		if ( points.length > 1 ) {
			x /= points.length;
			y /= points.length;
		}
		return new Vec2(x,y);
	}

	// A static method that returns true if the given point (an instance of Vec2) is inside the given polygon (an array of Vec2)
	static isPointInsidePolygon( q, polygonPoints ) {
		// This code was copied, with minor changes, from
		//    http://local.wasp.uwa.edu.au/~pbourke/geometry/insidepoly/
		// where it (the code, not the algorithm) is attributed to Randolph Franklin.
		// The idea behind the algorithm is to imagine a ray projecting
		// from the point toward the right, and then count how many times
		// that ray intersects an edge of the polygon.
		// If the number is odd, the point is inside the polygon.

		if ( polygonPoints.length < 3 )
			return false;

		let qx = q.x;
		let qy = q.y;
		let returnValue = false;
		let i = 0;
		let j = polygonPoints.length-1;
		while ( i < polygonPoints.length ) {
			let pi = polygonPoints[i];
			let xi = pi.x;
			let yi = pi.y;
			let pj = polygonPoints[j];
			let xj = pj.x;
			let yj = pj.y;

			if (
				(((yi <= qy) && (qy < yj)) || ((yj <= qy) && (qy < yi)))
				&& (qx < (xj - xi) * (qy - yi) / (yj - yi) + xi)
			) {
				returnValue = ! returnValue;
			}

			j = i;
			i ++;
		}
		return returnValue;
	}

	// Input: an array of Vec2
	// Return value: convex hull as an array of Vec2
	// Andrew's monotone chain convex hull algorithm
	// Adapted from https://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain
	static convexHull( input_points ) {

		// Returns the z-component of the cross product of vectors oa and ob.
		// Note that the x- and y-components of the cross product are zero,
		// because the z-components of oa and ob are both zero.
		// If this function returns a positive value, then ob is on the "left side" of oa
		// (assuming a right-handed coordinate system where we are looking down from the z+ axis).
		let cross = function(o,a,b) {
			return (a.x - o.x) * (b.y - o.y) - (a.y - o.y) * (b.x - o.x);
		};

		let points = input_points.slice(); // copies the array, to leave the caller's array untouched

		// Sort points so that those with a bigger x come later.
		// If two points have the same x, the one with the bigger y should come later.
		points.sort(function(a, b) {
			return a.x == b.x ? a.y - b.y : a.x - b.x;
		});

		// Find the lower half of the convex hull. "Lower" means the half on the y- side.
		let lower = [];
		let i;
		for (i = 0; i < points.length; i++) {
			while (lower.length >= 2 && cross(lower[lower.length - 2], lower[lower.length - 1], points[i]) <= 0) {
				// The vectors turn to the right, so get rid of the last point.
				lower.pop();
			}
			lower.push(points[i]);
		}

		// Find the upper half of the convex hull. "Upper" means the half on the y+ side.
		let upper = [];
		for (i = points.length - 1; i >= 0; i--) {
			while (upper.length >= 2 && cross(upper[upper.length - 2], upper[upper.length - 1], points[i]) <= 0) {
				// The vectors turn to the right, so get rid of the last point.
				upper.pop();
			}
			upper.push(points[i]);
		}

		// Get rid of redundant points.
		upper.pop();
		lower.pop();

		return lower.concat(upper);
	}

	// Input: a polygon, as an array of Vec2
	// Return value: a larger polygon, as an array of Vec2
	static computeExpandedPolygon( input_points, marginThickness ) {
		let newPoints = [];
		let p;
		if ( input_points.length === 0 ) {
			// do nothing
		}
		else if ( input_points.length === 1 ) {
			p = input_points[0];
			newPoints.push( new Vec2( p.x-marginThickness, p.y ) );
			newPoints.push( new Vec2( p.x, p.y-marginThickness ) );
			newPoints.push( new Vec2( p.x+marginThickness, p.y ) );
			newPoints.push( new Vec2( p.x, p.y+marginThickness ) );
		}
		else if ( input_points.length === 2 ) {
			let p0 = input_points[0];
			let p1 = input_points[1];
			let v0 = Vec2.mult( Vec2.diff(p1,p0).normalize(), marginThickness );
			let v1 = new Vec2( -v0.y, v0.x ); // a perpendicular vector
			newPoints.push( Vec2.sum( p0, v1 ) );
			newPoints.push( Vec2.sum( p0, v0.negate() ) );
			newPoints.push( Vec2.sum( p0, v1.negate() ) );
			newPoints.push( Vec2.sum( p1, v1.negate() ) );
			newPoints.push( Vec2.sum( p1, v0 ) );
			newPoints.push( Vec2.sum( p1, v1 ) );
		}
		else {
			for ( let i = 0; i < input_points.length; ++i ) {
				p = input_points[i];
				let p_previous = input_points[ i===0 ? input_points.length-1 : i-1 ];
				let p_next = input_points[ (i+1) % input_points.length ];
				let v_previous = Vec2.diff( p, p_previous ).normalize();
				let v_next = Vec2.diff( p, p_next ).normalize();

				newPoints.push( Vec2.sum( p, Vec2.mult(new Vec2(v_previous.y,-v_previous.x),marginThickness) ) );
				newPoints.push( Vec2.sum( p, Vec2.mult( Vec2.sum(v_next,v_previous).normalize(), marginThickness ) ) );
				newPoints.push( Vec2.sum( p, Vec2.mult(new Vec2(-v_next.y,v_next.x),marginThickness) ) );
			}
		}
		return newPoints;
	}

	// Returns null if the PCA fails.
	// Otherwise, returns on object { eigenvector1, eigenvector2, eigenvalue1, eigenvalue2 }
	// where eigenvector1 is the principle eigenvector,
	// both vectors are instances of Vec2, and both eigenvalues are simply numbers.
	static principleComponentAnalysis( points ) {
		if ( points.length < 2 )
			return null;

		// Compute the mean of the points
		let centroid = Vec2.centroidOfPoints( points );
		let meanX = centroid.x;
		let meanY = centroid.y;

		// Compute the covariance matrix (which is a symmetric 2x2 matrix):
		//
		//   [  covXX  covXY  ]
		//   [  covXY  covYY  ]
		//
		// where "covAB" is the covariance of A and B,
		// and "covAA" is the same as the variance of A.
		let covXX = 0;
		let covXY = 0;
		let covYY = 0;
		for ( let i = 0; i < points.length; ++i ) {
			let dx = points[i].x - meanX;
			let dy = points[i].y - meanY;
			covXX += dx * dx;
			covXY += dx * dy;
			covYY += dy * dy;
		}
		covXX /= ( points.length - 1 );
		covXY /= ( points.length - 1 );
		covYY /= ( points.length - 1 );

		// BEGIN: Perform eigendecomposition of the covariance matrix

		let discriminant = Math.sqrt( (covXX-covYY)*(covXX-covYY) + 4*covXY*covXY );
		// Note that, if the discriminant is nearly zero,
		// this means that the variance in X and in Y are nearly the same,
		// and that X and Y are nearly uncorrelated.
		// In other words, there's no single dominant direction for the data.
		let SMALL_POSITIVE_VALUE = 1e-8;
		if ( discriminant < SMALL_POSITIVE_VALUE )
			return null;

		let eVal1 = ((covXX+covYY) + discriminant)/2;
		let eVal2 = ((covXX+covYY) - discriminant)/2;

		// Note that, if eigenvalue1 is nearly zero,
		// then there's nearly no variance in the data in any direction,
		// and eigenvalue2 should be nearly zero as well.
		// However, if only eigenvalue2 is nearly zero,
		// then covXX*covYY is nearly equal to covXY^2, but that's not a problem.
		if ( Math.abs(eVal1) <= SMALL_POSITIVE_VALUE )
			return null;

		let eVec1 = new Vec2( covXY, eVal1-covXX ).normalize();
		let eVec2 = new Vec2( covXY, eVal2-covXX ).normalize();

		// END: Perform eigendecomposition of the covariance matrix

		return { eigenvector1 : eVec1, eigenvector2 : eVec2, eigenvalue1 : eVal1, eigenvalue2 : eVal2 };
	}

}

// ============================================================
// Box2 objects are for storing 2D axis-aligned rectangles.
// To create a new instance, use the new keyword:
//    let box_a = new Box2();
//    let box_b = new Box2(new Vec2(-10,-10),new Vec2(10,10));
// ============================================================

class Box2 {
	constructor( vec2_min = null, vec2_max = null ) {
		// Internally, the min and max points are diagonally opposite,
		// and are only valid if isEmpty===false.
		// Below, we initialize things based on what the client passed in.
		this.isEmpty = true;
		this.min = new Vec2();
		this.max = new Vec2();
		if ( vec2_min !== null && vec2_max !== null ) {
			this.boundPoint( vec2_min );
			this.boundPoint( vec2_max );
		}
	}
	clear() { this.isEmpty = true; this.min = new Vec2(); this.max = new Vec2(); }
	center() { return Vec2.average(this.min,this.max); }
	diagonal() { return Vec2.diff(this.max,this.min); }
	width() { return this.max.x - this.min.x; }
	height() { return this.max.y - this.min.y; }

	containsPoint( q ) {
		return !( this.isEmpty || q.x < this.min.x || q.x > this.max.x || q.y < this.min.y || q.y > this.max.y );
	}
	containsBox( b ) {
		if ( this.isEmpty ) return false;
		if ( b.isEmpty ) return true;
		return this.min.x <= b.min.x && b.max.x <= this.max.x && this.min.y <= b.min.y && b.max.y <= this.max.y;
	}

	// Enlarges the box enough to contain the given point
	boundPoint( vec2 ) {
		if ( this.isEmpty ) {
			this.isEmpty = false;
			this.min.copy( vec2 );
			this.max.copy( vec2 );
		}
		else {
			if ( vec2.x < this.min.x ) this.min.x = vec2.x;
			else if ( vec2.x > this.max.x ) this.max.x = vec2.x;

			if ( vec2.y < this.min.y ) this.min.y = vec2.y;
			else if ( vec2.y > this.max.y ) this.max.y = vec2.y;
		}
	}
	boundPoints( points ) {
		for ( let i = 0; i < points.length; ++i ) {
			this.boundPoint( points[i] );
		}
	}

	// Enlarges the box enough to contain the given box
	boundBox( box ) {
		if ( ! box.isEmpty ) {
			this.boundPoint( box.min );
			this.boundPoint( box.max );
		}
	}

}


